"""
These decorators are used to configure tests for
IDOR (Insecure Direct Object Reference) checks. We need to define two
different accounts (victim and attacker), so that we can check to see
whether the attacker is able to do actions that should only be limited
to the victim.
"""
from typing import Any
from typing import Callable
from typing import Dict

from ..datastore import inject_user_defined_variables
from .abstraction import get_abstraction


def victim_account(
    func: Callable[..., Dict[str, Any]],
) -> Callable[..., Dict[str, Any]]:
    """
    Example Usage:
        >>> @fuzz_lightyear.victim_account
        ... def victim_factory():
        ...     return {
        ...         'headers': {
        ...             'session': 'victim_session_id',
        ...         },
        ...     }
    """
    get_abstraction().get_victim_session = inject_user_defined_variables(func)
    return func


def attacker_account(
    func: Callable[..., Dict[str, Any]],
) -> Callable[..., Dict[str, Any]]:
    get_abstraction().get_attacker_session = inject_user_defined_variables(func)
    return func



from enum import Enum

from fuzz_lightyear.settings import get_settings


class AnsiColor(Enum):
    RESET = '[0m'
    BOLD = '[1m'
    RED = '[91m'
    RED_BACKGROUND = '[41m'
    YELLOW = '[33m'
    LIGHT_GREEN = '[92m'
    PURPLE = '[95m'


def colorize(text: str, color: AnsiColor) -> str:
    if not get_settings().enable_color:
        return text

    return '\x1b{}{}\x1b{}'.format(
        color.value,
        text,
        AnsiColor.RESET.value,
    )



import json
import string
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Tuple

import hypothesis.strategies as st
from hypothesis.strategies import SearchStrategy
from swagger_spec_validator.common import SwaggerValidationError    # type: ignore

from .datastore import get_user_defined_mapping
from .output.logging import log
from .settings import get_settings


def fuzz_parameters(
    parameters: List[Tuple[str, Dict[str, Any]]],
) -> SearchStrategy:
    output = {}
    for name, parameter in parameters:
        output[name] = _fuzz_parameter(parameter)

    return st.fixed_dictionaries(output)


def _fuzz_parameter(
    parameter: Dict[str, Any],
    required: bool = False,
) -> SearchStrategy:
    """
    :param required: for object types, the required parameter is in a
        separate array, rather than being attached to each parameter
        object. This parameter allows objects to pass in this information.
        e.g. {
            'type': 'object',
            'required': [
                'name',
            ],
            'properties': {
                'name': {
                    'type': 'string',
                },
            },
        }
    """
    required = parameter.get('required', required)

    _type = parameter.get('type')
    if not _type:
        raise SwaggerValidationError(
            'Missing \'type\' from {}'.format(
                json.dumps(parameter),
            ),
        )

    strategy = _get_strategy_from_factory(_type, parameter.get('name'))

    if not strategy:
        if 'enum' in parameter:
            return st.sampled_from(parameter['enum'])

        # As per https://swagger.io/docs/specification/data-models/data-types,
        # there are only a limited set of data types.
        mapping = {     # type: ignore # mypy doesn't like dynamic function signatures
            'string': _fuzz_string,
            'number': _fuzz_number,
            'integer': _fuzz_integer,
            'boolean': _fuzz_boolean,
            'array': _fuzz_array,
            'object': _fuzz_object,

            # TODO: handle `file` type
            # https://swagger.io/docs/specification/2-0/file-upload/
        }
        strategy = mapping[_type](parameter, required=required)

    # NOTE: We don't currently support `nullable` values, so we use `None` as a
    #       proxy to exclude the parameter from the final dictionary.
    if (
        # `name` check is used here as a heuristic to determine whether in
        # recursive call (arrays).
        parameter.get('name')
        and not required
    ):
        return st.one_of(st.none(), strategy)
    return strategy


def _fuzz_string(
    parameter: Dict[str, Any],
    required: bool = False,
) -> SearchStrategy:
    if parameter.get('in', None) == 'header':
        return st.text(
            # According to RFC 7230, non-ascii letters are deprecated, and there's
            # no telling what the server will do if they are sent. Since the intent
            # is not to break the server, but to send valid requests instead, we're
            # just going to limit it accordingly.
            alphabet=string.ascii_letters,
        )

    # TODO: Handle a bunch of swagger string formats.
    # https://swagger.io/docs/specification/data-models/data-types/#string
    kwargs = {}                                     # type: Dict[str, Any]

    if parameter.get('required', required):
        kwargs['min_size'] = 1
    if not get_settings().unicode_enabled:
        kwargs['alphabet'] = string.printable

    return st.text(**kwargs)


def _find_bounds(schema: Dict[str, Any]) -> Dict[str, Any]:
    """If the schema specifies a maximum and minimum for the numeric value,
       record it.
       By default, maximums and minimums in swagger are not exclusive"""
    bounds = {}
    if 'minimum' in schema:
        bounds['min_value'] = schema['minimum']
        if schema.get('exclusiveMinimum'):
            bounds['min_vallue'] += 1.0

    if 'maximum' in schema:
        bounds['max_value'] = schema['maximum']
        if schema.get('exclusiveMaximum'):
            bounds['max_value'] -= 1.0
    return bounds


def _fuzz_number(
    parameter: Dict[str, Any],
    **kwargs: Any,
) -> SearchStrategy:
    # TODO: Handle all the optional qualifiers for numbers.
    # https://swagger.io/docs/specification/data-models/data-types/#numbers
    bounds = _find_bounds(parameter)

    return st.floats(**bounds)


def _fuzz_integer(
    parameter: Dict[str, Any],
    **kwargs: Any,
) -> SearchStrategy:
    # TODO: Handle all the optional qualifiers for numbers.
    # https://swagger.io/docs/specification/data-models/data-types/#numbers
    bounds = _find_bounds(parameter)

    return st.integers(**bounds)


def _fuzz_boolean(
    parameter: Dict[str, Any],
    **kwargs: Any,
) -> SearchStrategy:
    return st.booleans()


def _fuzz_array(
    parameter: Dict[str, Any],
    required: bool = False,
) -> SearchStrategy:
    item = parameter['items']
    required = parameter.get('required', required)

    # TODO: Handle `oneOf`
    strategy = st.lists(
        elements=_fuzz_parameter(item, required=required),
        min_size=parameter.get(
            'minItems',
            0 if not required else 1,
        ),
        max_size=parameter.get('maxItems', None),
    )
    if not required:
        return st.one_of(st.none(), strategy)

    return strategy


def _fuzz_object(
    parameter: Dict[str, Any],
    **kwargs: Any,
) -> SearchStrategy:
    # TODO: Handle `additionalProperties`
    output = {}
    for name, specification in parameter['properties'].items():
        try:
            strategy = _get_strategy_from_factory(specification['type'], name)
        except KeyError:
            log.error(
                'Invalid swagger specification: expected \'type\'. Got \'{}\''.format(
                    json.dumps(specification),
                ),
            )
            raise

        if strategy:
            output[name] = strategy
            continue

        # `required` can be True, False, or an array of names that
        # are required.
        required = parameter.get('required', False)
        if required and isinstance(required, list):
            required = name in required

        output[name] = _fuzz_parameter(
            specification,
            bool(required),
        )

    return st.fixed_dictionaries(output)


def _get_strategy_from_factory(
    expected_type: str,
    name: Optional[str] = None,
) -> Optional[SearchStrategy[Any]]:
    if name not in get_user_defined_mapping():
        return None

    def type_cast() -> Any:
        """Use known types to cast output, if applicable."""
        output = get_user_defined_mapping()[name]()
        if output is None:
            # NOTE: We don't currently support `nullable` values, so we use `None`
            #       as a proxy to exclude the parameter from the final dictionary.
            return None
        if expected_type == 'string':
            return str(output)
        elif expected_type == 'integer':
            return int(output)

        return output

    return st.builds(type_cast)



import json
from collections import defaultdict
from functools import lru_cache
from typing import Any
from typing import Callable
from typing import cast
from typing import Dict
from typing import List
from typing import Optional
from urllib.parse import quote_plus
from urllib.parse import urlencode

from bravado.client import CallableOperation
from bravado_core.param import get_param_type_spec      # type: ignore
from cached_property import cached_property             # type: ignore
from hypothesis.strategies import SearchStrategy

from .datastore import get_post_fuzz_hooks
from .fuzzer import fuzz_parameters
from .output.logging import log
from .output.util import print_warning
from .supplements.abstraction import get_abstraction

COLLECTION_FORMAT_CHARS = {
    'csv': ',',
    'tsv': '\t',
    'pipes': '|',
    'ssv': ' ',
}


class FuzzingRequest:
    def __init__(
        self,
        operation_id: str,
        tag: str = 'default',
        **kwargs: Any,
    ) -> None:
        """
        :param operation_id: unique identifier for each Swagger operation.
        :param tag: this is how Swagger operations are grouped.
        """
        self.tag = tag
        self.operation_id = operation_id

        self.fuzzed_input = kwargs              # type: Optional[Dict[str, Any]]
        if not self.fuzzed_input:
            self.fuzzed_input = None

        # This SearchStrategy should be generated with hypothesis' `fixed_dictionaries`,
        # mapping keys to SearchStrategy.
        self._fuzzed_input_factory = None       # type: Optional[SearchStrategy]

    @property
    def id(self) -> str:
        return '{}.{}'.format(
            self.tag,
            self.operation_id,
        )

    def _encode_array_in_path(
        self,
        fuzzed_input: List,
        collection_format: str,
    ) -> str:
        separator = quote_plus(COLLECTION_FORMAT_CHARS[collection_format])
        return separator.join([str(i) for i in fuzzed_input])

    def json(self) -> Dict[str, Any]:
        path = self._swagger_operation.path_name    # type: str
        params = defaultdict(dict)                  # type: Dict[str, Dict[str, Any]]
        if self.fuzzed_input:
            for key, value in self._swagger_operation.params.items():
                if key not in self.fuzzed_input:
                    continue

                if value.location == 'path':
                    if value.param_spec['type'] == 'array':
                        fuzzed_input = self._encode_array_in_path(
                            self.fuzzed_input[key],
                            value.param_spec.get('collectionFormat', 'csv'),
                        )
                    else:
                        fuzzed_input = str(self.fuzzed_input[key])
                    path = path.replace(
                        f'{{{key}}}',
                        fuzzed_input,
                    )
                else:
                    params[value.location][key] = self.fuzzed_input[key]

        return {
            'method': self._swagger_operation.http_method.upper(),
            'path': path,
            **params,
        }

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}({self.tag}.{self.operation_id})'

    def __str__(self) -> str:
        data = self.json()
        url = (
            f'{self._swagger_operation.swagger_spec.api_url.rstrip("/")}'
            f'{data["path"]}'
        )

        if 'query' in data:
            url += '?'
            for key, value in data['query'].items():
                if not isinstance(value, list):
                    # NOTE: value should not be a dict, for a query param.
                    value = [value]
                for v in value:
                    url += f'{key}={quote_plus(str(v).encode())}&'

            url = url.rstrip('&')

        args = []
        if 'formData' in data:
            args.append(f'--data \'{urlencode(data["formData"])}\'')

        if 'body' in data:
            # NOTE: The first `body` key specifies where it's found.
            #       The second specifies the object that was fuzzed.
            #       Since there's only *one* body parameter (with a variable
            #       name), we can parse it out as such.
            args.append(f'--data \'{json.dumps(list(data["body"].values())[0])}\'')
            data['header'] = data.get('header', {})
            data['header']['Content-Type'] = 'application/json'

        # NOTE: We default to using the victim's auth headers, assuming that
        # this function is primarily used for easy reproduction (in which,
        # it should not matter the specific session that we use).
        headers = data.get('header', {})
        headers.update(
            get_victim_session_factory()().get(
                '_request_options', {},
            ).get(
                'headers', {},
            ),
        )
        for key, value in headers.items():
            args.append(f'-H \'{key}: {value}\'')

        return f'curl -X {data["method"]} {url} {" ".join(args)}'.rstrip()

    def send(
        self,
        auth: Optional[Dict[str, Any]] = None,
        *args: Any,
        should_log: bool = True,
        data: Optional[Dict[str, Any]] = None,
        **kwargs: Any,
    ) -> Any:
        """
        :param auth: parameters to pass to abstracted request method to specify
            the user making the request.
        :param should_log: this should only be false, if we're sending a
            duplicate request as part of a plugin.
        """
        if not data:
            data = {}

        # Empty dictionary means we're not sending parameters.
        if self.fuzzed_input is None:
            self.fuzzed_input = self.fuzz(data)
            self.apply_post_fuzz_hooks(self.fuzzed_input, rerun=False)
        else:
            # On the first run, self.fuzzed_input will be initialized
            # to a dictionary. Therefore, if this request is sent again,
            # this code path will execute, and we will want to rerun
            # the applicable post_fuzz hooks.
            #
            # TODO: I wonder whether we have to distinguish between
            # explicitly initialized `fuzzed_input`, and *actually* fuzzed
            # input, given that the former case will technically be the
            # *first* time the post_fuzz hook is run.
            self.apply_post_fuzz_hooks(self.fuzzed_input, rerun=True)

        if not auth:
            auth = get_victim_session_factory()()

        if should_log:
            log.info(str(self))

        _merge_auth_headers(self.fuzzed_input, auth)

        # auth details should override fuzzed_input, because
        # specifics should always override randomly generated content
        kwargs = _merge_kwargs(self.fuzzed_input, auth, kwargs)

        return get_abstraction().request_method(
            operation_id=self.operation_id,
            tag=self.tag,
            *args,
            **kwargs,
        )

    def fuzz(self, existing_data: Dict[str, Any]) -> Dict[str, Any]:
        """Returns a dictionary of values which can be used
        to call the operation being fuzzed.
        """
        if not self._fuzzed_input_factory:
            parameters = []
            for name, param in self._swagger_operation.params.items():
                specification = get_param_type_spec(param).copy()
                if param.location == 'body':
                    # For 'body' parameters, bravado discards information from the
                    # param spec itself. We pass in the 'required' parameter in this
                    # case.
                    # For the 'name' argument (seeing that body parameters can be
                    # named differently), we pass it in separately as it breaks the
                    # swagger specification if we group it together.
                    specification['required'] = param.required

                parameters.append((name, specification,))

            self._fuzzed_input_factory = fuzz_parameters(parameters)

        # NOTE: If we were really worried about performance later on,
        #       we might be able to address this. Specifically, we don't
        #       *need* to generate examples, just to throw it away later
        #       if the key is already in data.
        #       However, this involves parameter modification, which may
        #       require a more involved change.
        fuzzed_input = {}
        for key, value in self._fuzzed_input_factory.example().items():
            if key in existing_data:
                fuzzed_input[key] = existing_data[key]
                continue

            if value is not None:
                fuzzed_input[key] = value

        return fuzzed_input

    def apply_post_fuzz_hooks(
        self,
        fuzzed_input: Dict[str, Any],
        rerun: Optional[bool] = False,
    ) -> None:
        """After parameters for a request are fuzzed, this function
        applies developer-supplied post-fuzz hooks to the fuzzed
        input.

        :param fuzzed_input: The initial fuzz result from `self.fuzz`.
        """
        hooks = get_post_fuzz_hooks(self.operation_id, self.tag, rerun)
        for hook in hooks:
            hook(
                self._swagger_operation,
                fuzzed_input,
            )

    @cached_property        # type: ignore
    def _swagger_operation(self) -> CallableOperation:
        return cast(
            CallableOperation,
            getattr(
                getattr(get_abstraction().client, self.tag),
                self.operation_id,
            ),
        )


@lru_cache(maxsize=1)
def get_victim_session_factory() -> Callable[..., Dict[str, Any]]:
    factory = get_abstraction().get_victim_session
    if factory:
        return factory

    print_warning('No auth method specified.')
    return lambda: {}


def _merge_auth_headers(fuzzed_params: Dict[str, Any], auth: Dict[str, Any]) -> None:
    """
    The underlying Bravado client allows us to specify request headers on a per-request
    basis (https://bravado.readthedocs.io/en/stable/configuration.html#per-request-configuration).  # noqa: E501
    However, when there are authorization headers specified by the Swagger specification,
    the Bravado library parses it and merges it within the parameters required for the
    callable operation.

    This means, our fuzzing engine will set a value for it, which will override the
    manually specified authorization headers. To address this fact, we replace the fuzzed
    header with the actual one.
    """
    if not auth.get('_request_options', {}).get('headers', None):
        return

    for key in auth['_request_options']['headers']:
        # It looks like Bravado does some serialization for Python purposes, so we need
        # to mirror this.
        key = key.replace('-', '_')
        if key in fuzzed_params:
            fuzzed_params.pop(key)


def _merge_kwargs(*args: Any) -> Dict[str, Any]:
    """Merges the input dictionaries into a single dictionary which
    can be used in a fuzzing request.

    We need to merge headers and request_options seperately since
    they themselves are dictionaries.
    """

    headers = {}  # type: Dict[str, str]
    for dictionary in args:
        headers.update(dictionary.get('_request_options', {}).get('headers', {}))

    request_options = {}  # type: Dict[str, Any]
    for dictionary in args:
        # NOTE: this code does not merge the Bravado
        # response_callbacks option correctly, but we don't use it
        # in fuzz-lightyear. _request_options docs:
        # https://bravado.readthedocs.io/en/stable/configuration.html
        request_options.update(dictionary.get('_request_options', {}))

    output = {}  # type: Dict[str, Any]
    for dictionary in args:
        output.update(dictionary)

    output['_request_options'] = request_options
    output['_request_options']['headers'] = headers

    return output



from typing import Callable

from fuzz_lightyear.datastore import get_setup_fixtures


def setup(func: Callable) -> Callable:
    """
    Use the @setup decorator to mark functions that you'd like execute prior
    to running fuzz-lightyear tests

    Decorated functions should be in the file along with your factory fixtures

    Basic use:

        >>> import fuzz lightyear
        >>> @fuzz_lightyear.setup
        ... def setup_config():
        ...     // do configuration here

    """
    setup_fixtures = get_setup_fixtures()
    setup_fixtures.append(func)
    return func



from collections import defaultdict
from functools import lru_cache
from typing import Dict
from typing import Iterator
from typing import List
from typing import Optional
from typing import Tuple

from bravado.client import SwaggerClient

from .datastore import get_excluded_operations
from .datastore import get_included_tags
from .output.util import print_warning
from .request import FuzzingRequest
from .result import FuzzingResult
from .supplements.abstraction import get_abstraction


def get_fuzzable_tags(client: Optional[SwaggerClient]) -> List[str]:
    """Given a Swagger client, returns a list of tags that should
    actually be fuzzed. This respects the user-defined allowlist for
    tags.

    :param client: The Swagger client being fuzzed.
    :returns: A list of tags to fuzz.
    """
    if not client:
        return []

    allowlisted_tags = get_included_tags()
    if not allowlisted_tags:
        return dir(client)

    fuzzable_tags = []
    for tag in allowlisted_tags:
        if tag not in dir(client):
            print_warning(
                f'The tag "{tag}" is not in the Swagger schema, will not fuzz it.',
            )
            continue

        fuzzable_tags.append(tag)

    return fuzzable_tags


def generate_sequences(
    n: int,
    tests: Optional[Tuple[str, ...]] = None,
) -> Iterator[FuzzingResult]:
    """
    We perform DFS to obtain these sequences, since we want nice output.

    :param n: max length of each sequence.
    :param tests: specifies the tests we want to run.
    """
    # TODO: It would probably better if we can sort on ending operation
    #       (rather than starting operation), so that it's clearer for
    #       output.
    client = get_abstraction().client
    request_graph = _generate_request_graph()
    for tag_group in get_fuzzable_tags(client):
        last_results = []   # type: List
        for _ in range(n):
            good_sequences = []
            for sequence in _add_request_to_sequence(
                tag_group, last_results, request_graph,
            ):
                # TODO: We can probably modify this algorithm to be better.
                if (
                    tests and not
                    (
                        sequence[-1].id in tests
                        or sequence[-1].tag in tests
                    )
                ):
                    continue

                result = FuzzingResult(sequence)
                yield result

                if result.is_successful():
                    good_sequences.append(sequence)

            last_results = good_sequences


def _add_request_to_sequence(
    tag_group: str,
    seed: Optional[List[List[FuzzingRequest]]] = None,
    request_graph: Dict[str, set] = {},
) -> List[List[FuzzingRequest]]:
    """
    :returns output: list of request sequences
    """
    client = get_abstraction().client

    if not seed:
        return [
            [request]
            for request in _generate_requests(tag_group)
        ]

    output = []
    for sequence in seed:
        for tag_group in dir(client):
            for request in _generate_requests(tag_group):
                neighbor_requests = request_graph[request.operation_id]
                # We have no idea whether a request should consume an existing
                # resource, use a factory, or fuzz a value.
                #
                # However, we want to move towards building sequences of requests
                # that only consume resources created by previous requests (the
                # algorithm described in the original Microsoft paper)
                #
                # Therefore, add the request to the sequence if it shares an edge with
                # ANY of the requests in the sequence- leaving room for factory
                # usage, but still keeping in line with the spirit of RESTler.
                if any([req.operation_id in neighbor_requests for req in sequence]):
                    new_sequence = sequence + [request]
                    output.append(new_sequence)

    return output


@lru_cache(maxsize=1)
def _generate_request_graph() -> Dict[str, set]:
    """
    Generates a directed graph, represented as an adjacency list.

    Every node in the graph is an operation_id, and an edge between two nodes
    A and B means that node A consumes a resource that node B produces.

    """
    client = get_abstraction().client
    produces = defaultdict(list)  # type: Dict[str, List[str]]
    consumes = {}  # type: Dict[str, List[str]]

    for tag_group in dir(client):
        for operation_id in dir(getattr(client, tag_group)):
            operation = getattr(getattr(client, tag_group), operation_id).operation
            responses = operation.op_spec.get('responses', {}).get('200', {})
            if responses:
                response_params = list(
                    responses.get('schema', {}).get('properties', {}).keys(),
                )
                for response_param in response_params:
                    produces[response_param].append(operation_id)

            consumes[operation_id] = list(operation.params.keys())

    request_graph = defaultdict(set)  # type: Dict[str, set]

    # This would be a graph that maps producing operations -> to consuming operations
    # for consuming_op, params in consumes.items():
    #    for param in params:
    #        for producing_op in produces[param]:
    #            request_graph[producing_op].append(consuming_op)

    for consuming_op, params in consumes.items():
        for param in params:
            request_graph[consuming_op].update(produces[param])

    return request_graph


def _generate_requests(tag_group: str) -> Iterator[FuzzingRequest]:
    """
    Generates requests based on the client's Swagger specification.
    Excludes methods specifed by developers with the
    `@fuzz_lightyear.exclusions.operations` decorator.
    """
    client = get_abstraction().client
    excluded_operations = get_excluded_operations()
    for operation_id in dir(getattr(client, tag_group)):
        if operation_id in excluded_operations:
            expected_tag = excluded_operations[operation_id]
            if expected_tag is not None and expected_tag != tag_group:
                print_warning(
                    f'Excluded operation {operation_id} expected to have the '
                    f'tag {expected_tag}, but actually had tag {tag_group}. '
                    f'Going to exclude {operation_id} anyways.',
                )

            continue

        yield FuzzingRequest(
            operation_id=operation_id,
            tag=tag_group,
        )



from typing import Any
from typing import Dict
from typing import List

from .request import FuzzingRequest
from .response import ResponseSequence


class FuzzingResult:
    def __init__(
        self,
        request_sequence: List[FuzzingRequest],
    ):
        self.requests = request_sequence    # type: List[FuzzingRequest]
        self.responses = ResponseSequence()  # type: ResponseSequence

        self.log_output = ''
        self.exception_info = {}            # type: Dict[str, Any]

    def is_successful(self) -> bool:
        return bool(self.responses)



from typing import Callable
from typing import Iterable
from typing import Optional
from typing import Union

from fuzz_lightyear.datastore import register_post_fuzz_hook
from fuzz_lightyear.util import listify_decorator_args


def post_fuzz(
    operations: Optional[Union[str, Iterable[str]]] = None,
    tags: Optional[Union[str, Iterable[str]]] = None,
    rerun: bool = True,
) -> Callable:

    # These are renamed just to make mypy happy.
    operation_ids = listify_decorator_args(operations)
    _tags = listify_decorator_args(tags)

    def decorator(func: Callable) -> Callable:
        register_post_fuzz_hook(
            hook=func,
            operation_ids=operation_ids,
            tags=_tags,
            rerun=rerun,
        )
        return func

    return decorator



from typing import Iterable
from typing import List
from typing import Optional
from typing import Union


def listify_decorator_args(
    argument: Optional[Union[str, Iterable[str]]],
) -> List[str]:
    """Converts a user-supplied "list" to an actual Python list.
    This helper supports a "string" list as follows:
        >>> listify_decorator_args('a,b,c')
        ['a', 'b', 'c']

    :param argument: The argument to listify.
    :return: A list of strings..
    """
    if not argument:
        return []

    if isinstance(argument, str):
        return [
            arg.strip()
            for arg in argument.split(',')
        ]

    return list(argument)



from typing import Any
from typing import Dict
from typing import List

from .request import FuzzingRequest


class ResponseSequence:

    def __init__(self) -> None:
        # Collection of response objects returned by API calls.
        self.responses = []     # type: List[Any]

        # Collection of items produced through making requests
        self.data = {}          # type: Dict[str, Any]

        # Result of analysis done on responses received
        self.test_results = {}  # type: Dict[str, bool]

    def __len__(self) -> int:
        return len(self.responses)

    def add_response(self, response: Any) -> None:
        self.responses.append(response)
        if not response:
            return

        for key in dir(response):
            self.data[key] = getattr(response, key)

    def analyze_requests(
        self,
        request_sequence: List[FuzzingRequest],
    ) -> None:
        # We import here because otherwise we run into a circular dependency
        from .plugins import get_enabled_plugins
        plugins = get_enabled_plugins()
        for plugin in plugins:
            if plugin.should_run(
                request_sequence,
                self.responses,
            ):
                self.test_results[plugin.__name__] = plugin.is_vulnerable(
                    request_sequence,
                    self.responses,
                )



import importlib.util
import os
from types import ModuleType
from typing import Optional


def import_fixtures(path: str) -> None:
    """Recursively imports all python files from a path."""
    if os.path.isfile(path):
        import_module_from_path(path)
        return

    for root, _, filenames in os.walk(path):
        for filename in filenames:
            path = os.path.join(root, filename)
            if not path.endswith('.py'):
                continue

            import_module_from_path(path)


def import_module_from_path(path: str) -> Optional[ModuleType]:
    """
    Source: https://stackoverflow.com/a/67692

    :param path: path to python file to import
    """
    # I think this just needs to be unique.
    module_name = os.path.splitext(path)[0].replace('/', '.')

    spec = importlib.util.spec_from_file_location(
        module_name,
        path,
    )
    if not spec:
        return None

    module = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(module)     # type: ignore

    return module



from abc import ABCMeta
from typing import Any
from typing import List

from ..request import FuzzingRequest


class BasePlugin(metaclass=ABCMeta):

    @staticmethod
    def should_run(
        request_sequence: List[FuzzingRequest],
        response_sequence: List[Any],
    ) -> bool:
        return True

    @staticmethod
    def is_vulnerable(
        request_sequence: List[FuzzingRequest],
        response_sequence: List[Any],
    ) -> bool:
        raise NotImplementedError



import inspect
from collections import defaultdict
from functools import lru_cache
from functools import wraps
from typing import Any
from typing import Callable
from typing import Dict
from typing import List
from typing import Optional
from typing import Set
from typing import Tuple

from bravado.client import CallableOperation


PostFuzzHook = Callable[[CallableOperation, Dict[str, Any]], None]


# These are module variables which contain the post-fuzz hooks
# which have been registered. Each global allows fuzz_lightyear
# to get a list applicable to a certain operation or tag.
_ALL_POST_FUZZ_HOOKS_BY_OPERATION = defaultdict(set)  # type: Dict[str, Set[PostFuzzHook]]
_ALL_POST_FUZZ_HOOKS_BY_TAG = defaultdict(set)  # type: Dict[str, Set[PostFuzzHook]]

# This probably isn't scalable if we add more parameters to
# hooks , but for now this works, and is easy to understand.
_RERUN_POST_FUZZ_HOOKS_BY_OPERATION = defaultdict(set)  # type: Dict[str, Set[PostFuzzHook]]  # noqa: E501
_RERUN_POST_FUZZ_HOOKS_BY_TAG = defaultdict(set)  # type: Dict[str, Set[PostFuzzHook]]


def register_post_fuzz_hook(
    hook: PostFuzzHook,
    operation_ids: Optional[List[str]] = None,
    tags: Optional[List[str]] = None,
    rerun: bool = True,
) -> None:
    """Adds a post-fuzz hook to fuzz_lightyear's store of post-fuzz
    hooks.

    :param hook: The hook.
    :param operation_ids: A list of operations that the input hook
    applies to.
    :param tags: A list of Swagger tags that the input hook applies
    to.
    :param rerun: Whether this hook needs to be rerun if a request
    is rerun (for example, in the idor plugin).
    """
    if not operation_ids and not tags:
        operation_ids = ['*']

    if not operation_ids:
        operation_ids = []

    if not tags:
        tags = []

    for operation_id in operation_ids:
        _ALL_POST_FUZZ_HOOKS_BY_OPERATION[operation_id].add(hook)
        if rerun:
            _RERUN_POST_FUZZ_HOOKS_BY_OPERATION[operation_id].add(hook)

    for tag in tags:
        _ALL_POST_FUZZ_HOOKS_BY_TAG[tag].add(hook)
        if rerun:
            _RERUN_POST_FUZZ_HOOKS_BY_TAG[tag].add(hook)


def get_post_fuzz_hooks(
    operation_id: str,
    tag: Optional[str] = None,
    rerun: Optional[bool] = False,
) -> List[PostFuzzHook]:
    """Returns a list of functions that should be applied to fuzzed
    data for the input operation.
    """
    operation_hook_store = _RERUN_POST_FUZZ_HOOKS_BY_OPERATION \
        if rerun \
        else _ALL_POST_FUZZ_HOOKS_BY_OPERATION

    tag_hook_store = _RERUN_POST_FUZZ_HOOKS_BY_TAG \
        if rerun \
        else _ALL_POST_FUZZ_HOOKS_BY_TAG

    operation_hooks = operation_hook_store[operation_id].union(
        operation_hook_store['*'],
    )
    tag_hooks = tag_hook_store[tag] if tag else set()
    return list(operation_hooks.union(tag_hooks))


@lru_cache(maxsize=1)
def get_setup_fixtures() -> List:
    """
    This is a global list that contains the functions that should be executed
    before fuzz-lightyear begins executing tests.

    :rtype: list(function)
    """
    return []


@lru_cache(maxsize=1)
def get_user_defined_mapping() -> Dict:
    """
    This is essentially a global variable, within a function scope, because
    this returns a reference to the cached dictionary.

    :rtype: dict(str => function)
    """
    return {}


@lru_cache(maxsize=1)
def get_included_tags() -> Set[str]:
    """This is a global set containing tags which should
    be fuzzed. Each element is a string for the tag which
    should be included.
    """
    return set()


@lru_cache(maxsize=1)
def get_excluded_operations() -> Dict[str, Optional[str]]:
    """
    This is a global dictionary containing fuzzing-excluded operations.
    Operation id's are keys. Tags are values, if the user provided them.
    If you don't care about the operation's tag, you can get just the
    excluded operations with `get_excluded_operations().keys()`.

    :rtype: dict(str => str)
    """
    return {}


@lru_cache(maxsize=1)
def get_non_vulnerable_operations() -> Dict[str, Optional[str]]:
    """
    This is a global dictionary containing non-vulnerable operations.
    Operation ids are keys. Tags are values, if the user provided them.
    If you don't care about the operation's tag, you can get just the
    excluded operations with `get_excluded_operations().keys()`.

    :rtype: dict(str => str)
    """
    return {}


def clear_cache() -> None:
    """ Clear the cached values for fixture functions """
    for value in get_user_defined_mapping().values():
        value._fuzz_cache = None


def inject_user_defined_variables(func: Callable) -> Callable:
    """
    This decorator allows the use of user defined variables in functions.
    e.g.
        >>> @fuzz_lightyear.register_factory('name')
        ... def name():
        ...     return 'freddy'
        >>>
        >>> @inject_user_defined_variables
        ... def foobar(name):
        ...     print(name)     # freddy
    """
    mapping = get_user_defined_mapping()

    @wraps(func)
    def wrapped(*args: Any, **kwargs: Any) -> Any:
        if getattr(func, '_fuzz_cache', None) is not None:
            return func._fuzz_cache  # type: ignore

        expected_args = _get_injectable_variables(func)
        type_annotations = inspect.getfullargspec(func).annotations

        for index, arg_name in enumerate(expected_args):
            if index < len(args):
                # This handles the case of explicitly supplied
                # positional arguments, so that we don't pass func
                # two values for the same argument.
                continue

            if arg_name not in mapping:
                raise TypeError

            value = mapping[arg_name]()
            if (
                arg_name in type_annotations and
                not isinstance(type_annotations[arg_name], type(List))
            ):
                # If type annotations are used, use that to cast
                # values for input.
                value = type_annotations[arg_name](value)

            kwargs[arg_name] = value

        func._fuzz_cache = func(*args, **kwargs)  # type: ignore
        return func._fuzz_cache  # type: ignore

    return wrapped


def _get_injectable_variables(func: Callable) -> Tuple[str, ...]:
    """
    The easiest way to understand this is to see it as an example:

        >>> def func(a, b=1, *args, c, d=2, **kwargs):
        ...     e = 5
        >>>
        >>> print(func.__code__.co_varnames)
        ('a', 'b', 'c', 'd', 'args', 'kwargs', 'e')
        >>> print(func.__code__.co_argcount)    # `a` and `b`
        2
        >>> print(func.__code__.co_kwonlyargcount)  # `c` and `d`
        2
    """
    variable_names = func.__code__.co_varnames
    arg_count = func.__code__.co_argcount + func.__code__.co_kwonlyargcount

    return variable_names[:arg_count]



from typing import Any


class BaseFuzzingError(Exception):
    """This exists so that clients can distinguish between exception classes."""


class ConflictingKeys(BaseFuzzingError):
    def __init__(
        self,
        key: str,
        *args: Any,
    ) -> None:
        return super().__init__(
            'There are multiple factory registrations for "{}".'.format(key),
            *args,
        )


class ConflictingHandlers(BaseFuzzingError):
    def __init__(
        self,
        key: str,
        *args: Any,
    ) -> None:
        return super().__init__(
            'There are multiple handlers registered for "{}".'.format(key),
            *args,
        )



import random
from functools import lru_cache

from hypothesis import core


class Settings:
    def __init__(self) -> None:
        self.seed = random.getrandbits(128)    # type: int
        self.unicode_enabled = True            # type: bool
        self.enable_color = True               # type: bool

    @property
    def seed(self) -> int:
        return self._seed

    @seed.setter
    def seed(self, value: int) -> None:
        self._seed = value
        core.global_force_seed = value      # type: ignore
        random.seed(value)


@lru_cache(maxsize=1)
def get_settings() -> Settings:
    return Settings()



"""
Not all endpoints which allow a direct object reference needs to be
authenticated. The endpoint could simply be providing non-sensitive information
about the object being queried.

To address this, developers can specify operations which should not be used
in the fuzzing process. We provide two levels of exclusions: non-vulnerable
exclusions and general exclusions.

Non-vulnerable exclusions exclude operations which can be included in the
fuzzing process, but represent a low risk of vulnerabilities, and thus should
not be checked for vulnerabilities for speed of fuzzing and clarity of fuzzing
output.

General exclusions exclude operations which should not be included in the
fuzzing process. Operations that can go here can include operations which don't
function correctly in testing environments.
"""
from functools import wraps
from typing import Callable
from typing import Dict
from typing import Optional
from typing import Tuple

from fuzz_lightyear.datastore import get_excluded_operations
from fuzz_lightyear.datastore import get_non_vulnerable_operations
from fuzz_lightyear.output.util import print_warning


def non_vulnerable_operations(func: Callable) -> Callable:
    """Allows developers to specify operations which
    should not be tested for vulnerabilities.

    Examples:
        Ignoring operations specified by operation ids in lists
            >>> @fuzz_lightyear.exclude.non_vulnerable_operations
            ... def b():
            ...     return ['get_pets', 'get_store_inventory']

        Ignoring operations specified by "tag.operation_id" in lists
            >>> @fuzz_lightyear.exclude.non_vulnerable_operations
            ... def c():
                    return ['pets.get_pets', 'store.get_store_inventory']
    """
    get_operations_fn = _get_formatted_operations(func)
    get_non_vulnerable_operations().update(get_operations_fn())

    return func


def operations(func: Callable) -> Callable:
    """Allows developers to specify operations which
    should not be called in the fuzzing process.

    Examples:
        Ignoring operations specified by operation ids in lists
            >>> @fuzz_lightyear.exclude.operations
            ... def b():
            ...     return ['get_pets', 'get_store_inventory']

        Ignoring operations specified by "tag.operation_id" in lists
            >>> @fuzz_lightyear.exclude.operations
            ... def c():
            ...     return ['pets.get_pets', 'store.get_store_inventory']
    """
    get_operations_fn = _get_formatted_operations(func)
    get_excluded_operations().update(get_operations_fn())

    return func


def _get_formatted_operations(func: Callable) -> Callable[[], Dict[str, Optional[str]]]:
    """
    Given a user-defined function which specifies Swagger operations in a
    supported form,, returns a function `f`. `f()` returns a mapping
    from swagger operation id to its tag.
    """
    @wraps(func)
    def wrapped() -> Dict[str, Optional[str]]:
        operations = func()
        result = {}
        for operation in operations:
            formatted_operation = _format_operation(operation)
            if formatted_operation:
                operation_id, tag = formatted_operation
                result[operation_id] = tag

        return result

    return wrapped


def _format_operation(operation: str) -> Optional[Tuple[str, Optional[str]]]:
    if isinstance(operation, str):
        num_dots = operation.count('.')
        if num_dots == 0:
            return (operation, None)
        elif num_dots == 1:
            tag, operation_id = operation.split('.')
            return (operation_id, tag)

    print_warning(
        f'Failed to interpret {str(operation)} as an operation to exclude.',
    )
    return None



"""
Due to a micro-service ecosystem, services' swagger files may not come
with complete CRUD (Create-Retrieve-Update-Delete) functionality. This
makes it difficult to test since we lack the resource IDs necessary to
determine whether there are vulnerabilities in the retrieve/update/delete
functions.

In order to address this issue, we need to supplement our tests with
an ability to create adhoc resources directly. These functions support this,
and allow developers to configure their tests as necessary.
"""
from typing import Callable
from typing import Iterable
from typing import Union

from fuzz_lightyear.datastore import get_user_defined_mapping
from fuzz_lightyear.datastore import inject_user_defined_variables
from fuzz_lightyear.exceptions import ConflictingKeys
from fuzz_lightyear.util import listify_decorator_args


def register_factory(keys: Union[str, Iterable[str]]) -> Callable:
    """
    :type keys: str|iterable

    Basic Use:
        >>> import fuzz_lightyear
        >>> @fuzz_lightyear.register_factory('user_id')
        ... def create_user():
        ...     return 1

    Multiple Keys:
        >>> @fuzz_lightyear.register_factory(['biz_id', 'business_id'])
        ... def create_business():
        ...     return 1
        >>>
        >>> @fuzz_lightyear.register_factory('biz_id, business_id')
        ... def create_business():
        ...     return 1

    Dependency Injection:
        >>> @fuzz_lightyear.register_factory('biz_id')
        ... def create_business():
        ...     return 1
        >>>
        >>> @fuzz_lightyear.register_factory('biz_user_id')
        ... def create_biz_user(biz_id):
        ...     assert biz_id == 1
        ...     return 2

    Type Hinting:
        This is free for root level factories, using the expected type based
        on the swagger specification. For transitive use of factories, we
        employ type-casting through type hints.

        I'm aware that this means Python 3 will be required (and may not be
        easy to implement for legacy codebases) but this is the easiest way
        to implement this (without essentially implementing it ourselves).

        For more edge cases and considerations, check out the test cases.

        >>> @fuzz_lightyear.register_factory('biz_id')
        ... def create_business():
        ...     return 3
        >>>
        >>> @fuzz_lightyear.register_factory('biz_user_id')
        ... def create_biz_user(biz_id: str):
        ...     assert biz_id == '3'
        ...     return 4
    """
    # This is renamed just to make mypy happy.
    _keys = listify_decorator_args(keys)

    def decorator(func: Callable) -> Callable:
        wrapped = inject_user_defined_variables(func)

        mapping = get_user_defined_mapping()
        for key in _keys:
            if key in mapping:
                raise ConflictingKeys(key)

            mapping[key] = wrapped

        return wrapped

    return decorator



from functools import lru_cache
from typing import Any
from typing import Callable
from typing import Optional

from bravado.client import SwaggerClient

from fuzz_lightyear.exceptions import ConflictingHandlers


class Abstraction:
    """
    Allows clients to configure various parts of the fuzzer's core functionality,
    all exposed through one single interface.
    """

    def __init__(self) -> None:
        self.get_victim_session = None      # type: Optional[Callable]
        self.get_attacker_session = None    # type: Optional[Callable]

        self.client = None                  # type: Optional[SwaggerClient]
        self._request_method = None         # type: Optional[Callable]

    @property
    def request_method(self) -> Callable:
        if self._request_method:
            return self._request_method

        return default_request_method

    @request_method.setter
    def request_method(self, func: Callable) -> None:
        if self._request_method:
            raise ConflictingHandlers('make_request')

        self._request_method = func


@lru_cache(maxsize=1)
def get_abstraction() -> Abstraction:
    return Abstraction()


def default_request_method(
    operation_id: str,
    tag: str = 'default',
    *args: Any,
    **kwargs: Any,
) -> Any:
    """
    :param operation_id: there's a unique operationId for each
        (tag, operation) in the Swagger schema

    :param tag: Swagger tag
    """
    future = getattr(
        getattr(get_abstraction().client, tag),
        operation_id,
    )(*args, **kwargs)
    return future.result()



from typing import Callable

from .abstraction import get_abstraction


def make_request(func: Callable) -> Callable:
    """
    Allows clients to customize how they want to make a request to the
    service undergoing fuzz testing.
    """
    get_abstraction().request_method = func
    return func


def custom_swagger_client(func: Callable) -> None:
    """
    Allows client to customize a SwaggerClient, so that they can leverage
    the default request handler.
    """
    get_abstraction().client = func()



from functools import lru_cache
from typing import Sequence
from typing import Type

from ..output.logging import log
from ..supplements.abstraction import get_abstraction
from .base import BasePlugin
from .idor import IDORPlugin


@lru_cache(maxsize=1)
def get_enabled_plugins() -> Sequence[Type[BasePlugin]]:
    plugins = []

    if (
        get_abstraction().get_victim_session and
        get_abstraction().get_attacker_session
    ):
        plugins.append(IDORPlugin)

    if not plugins:
        log.warning(
            'No plugins enabled! Be sure to check your configurations.',
        )

    return plugins



import warnings
from typing import Any
from typing import Dict
from typing import List
from typing import Optional

import requests
import simplejson
import yaml
from bravado.client import SwaggerClient
from bravado.exception import HTTPError
from hypothesis.errors import NonInteractiveExampleWarning
from swagger_spec_validator.common import SwaggerValidationError    # type: ignore

from .datastore import get_excluded_operations
from .datastore import get_non_vulnerable_operations
from .datastore import get_setup_fixtures
from .discovery import import_fixtures
from .generator import generate_sequences
from .output.interface import ResultFormatter
from .output.logging import log
from .output.util import print_error
from .runner import validate_sequence
from .settings import get_settings
from .supplements.abstraction import get_abstraction
from .usage import parse_args


def main(argv: Optional[List[str]] = None) -> int:
    args = parse_args(argv)
    if args.verbose:    # pragma: no cover
        log.set_debug_level(args.verbose)

    # NOTE: Hypothesis warns us against using `.example()` (which we leverage
    # during the fuzzing process), and to use `@given` instead. However, we
    # are not using hypothesis in a conventional manner, and therefore this
    # warning does not apply to us.
    warnings.filterwarnings('ignore', category=NonInteractiveExampleWarning)

    # Setup
    message = setup_client(args.url, args.schema)
    if message:
        print_error(message)
        return 1

    setup_fixtures(args.fixture)
    run_user_defined_setup()
    if args.ignore_non_vulnerable:
        get_excluded_operations().update(
            get_non_vulnerable_operations(),
        )

    outputter = run_tests(
        *args.test,
        iterations=args.iterations,
        seed=args.seed,
        ignore_exceptions=args.ignore_exceptions,
        disable_unicode=args.disable_unicode,
    )

    outputter.show_results()
    return outputter.stats['failure'] != 0


def setup_client(
    url: str,
    schema: Optional[Dict[str, Any]] = None,
) -> Optional[str]:
    """
    :returns: error message, if appropriate.
    """
    if get_abstraction().client:
        return None

    try:
        config = {'internally_dereference_refs': True}
        if not schema:
            client = SwaggerClient.from_url(url, config=config)
        else:
            client = SwaggerClient.from_spec(schema, origin_url=url, config=config)
    except requests.exceptions.ConnectionError:
        return 'Unable to connect to server.'
    except (
        simplejson.errors.JSONDecodeError,      # type: ignore
        yaml.YAMLError,
        HTTPError,
    ):
        return (
            'Invalid swagger file. Please check to make sure the '
            'swagger file can be found at: {}.'.format(url)
        )
    except SwaggerValidationError:
        return 'Invalid swagger format.'

    get_abstraction().client = client
    return None


def setup_fixtures(fixtures: List[str]) -> None:
    for fixture_path in fixtures:
        import_fixtures(fixture_path)


def run_user_defined_setup() -> None:
    for function in get_setup_fixtures():
        function()


def run_tests(
    *tests: str,
    iterations: int = 1,
    seed: int = None,
    ignore_exceptions: bool = False,
    disable_unicode: bool = False,
) -> ResultFormatter:
    """
    :param tests: list of tests to run.
        e.g. `basic.get_private_listing`

    :param iterations: size of request sequence to generate.
    :param seed: used for random generation of test input
    :param ignore_exceptions: if True, ignores HTTP exceptions to requests.
    :param disable_unicode: if True, only use ASCII characters to fuzz strings
    """
    if seed is not None:
        get_settings().seed = seed
    if disable_unicode:
        get_settings().unicode_enabled = False

    outputter = ResultFormatter()
    for result in generate_sequences(
        n=iterations,
        tests=tests,
    ):
        try:
            validate_sequence(result.requests, result.responses)
        except Exception as e:
            if (
                ignore_exceptions and
                isinstance(e, HTTPError)
            ):
                # This makes it look like a valid request sequence.
                result.responses.responses = result.requests
            else:
                outputter.record_exception(result, e)

        outputter.record_result(result)

    return outputter



import argparse
import json
import os
from typing import Any
from typing import cast
from typing import Dict
from typing import List
from typing import Optional

import yaml
try:
    from yaml import CSafeLoader as SafeLoader
except ImportError:
    from yaml import SafeLoader  # type: ignore


from fuzz_lightyear.version import VERSION


def parse_args(argv: Optional[List[str]] = None) -> argparse.Namespace:
    parser = argparse.ArgumentParser()
    parser.add_argument(
        '-v',
        '--verbose',
        help='Increase the verbosity of logging.',
        action='count',
    )
    parser.add_argument(
        '--version',
        action='version',
        version=VERSION,
        help='Displays version information.',
    )

    parser.add_argument(
        'url',
        type=str,
        help='URL of server to fuzz.',
    )
    parser.add_argument(
        '-n',
        '--iterations',
        nargs='?',
        default=1,
        help='Maximum request sequence length to fuzz.',
        type=int,
    )
    parser.add_argument(
        '--schema',
        help=(
            'Path to local swagger schema. If provided, this overrides the'
            'swagger file found at the URL.'
        ),
        type=_is_valid_schema,
    )

    parser.add_argument(
        '-f',
        '--fixture',
        action='append',
        default=[],
        help='Path to custom specified fixtures.',
        type=_is_valid_path,
    )

    parser.add_argument(
        '--seed',
        type=int,
        help='Specify seed for generation of random output.',
    )
    parser.add_argument(
        '-t',
        '--test',
        type=str,
        default=[],
        action='append',
        help='Specifies a single test to run.',
    )

    parser.add_argument(
        '--ignore-exceptions',
        action='store_true',
        help=(
            'Ignores all exceptions raised during fuzzing (aka. only fails when '
            'vulnerabilities are found).'
        ),
    )
    parser.add_argument(
        '--ignore-non-vulnerable',
        action='store_true',
        help=(
            'Excludes all non-vulnerable operations from tests. This is '
            'especially handy when you *only* want to check for '
            'vulnerabilities (rather than Swagger accuracy).'
        ),
    )

    parser.add_argument(
        '--disable-unicode',
        action='store_true',
        help=(
            'Disable fuzzing with unicode strings (only use ascii characters to '
            'fuzz strings).'
        ),
    )

    return parser.parse_args(argv)


def _is_valid_schema(path: str) -> Dict[str, Any]:
    _is_valid_path(path)

    if path.endswith('json'):
        return _is_json_schema(path)
    elif path.endswith(('yaml', 'yml')):
        return _is_yaml_schema(path)
    else:
        # try both json and yaml then raise if both file types
        # don't work
        try:
            return _is_json_schema(path)
        except argparse.ArgumentTypeError:
            pass

        try:
            return _is_yaml_schema(path)
        except argparse.ArgumentTypeError:
            pass

        raise argparse.ArgumentTypeError(
            'Invalid schema file: {}. Check that it is either JSON or YAML.'.format(path),
        )


def _is_json_schema(path: str) -> Dict[str, Any]:
    with open(path) as f:
        try:
            return cast(
                Dict[str, Any],
                json.loads(f.read()),
            )
        except json.decoder.JSONDecodeError:
            raise argparse.ArgumentTypeError(
                'Invalid JSON file: {}'.format(path),
            )


def _is_yaml_schema(path: str) -> Dict[str, Any]:
    with open(path) as f:
        try:
            return cast(
                Dict[str, Any],
                yaml.load(f, Loader=SafeLoader),
            )
        except yaml.YAMLError:
            raise argparse.ArgumentTypeError(
                'Invalid YAML file: {}'.format(path),
            )


def _is_valid_path(path: str) -> str:
    if not os.path.exists(path):
        raise argparse.ArgumentTypeError(
            'Invalid path: {}'.format(path),
        )

    return path



import sys
import textwrap
import traceback
from collections import Counter
from datetime import datetime
from typing import List
from typing import Optional

from . import formatter
from ..result import FuzzingResult
from ..settings import get_settings
from .color import AnsiColor
from .color import colorize
from .logging import log
from .logging import root_logger


class ResultFormatter:
    def __init__(self) -> None:
        if not sys.stdout.isatty():
            get_settings().enable_color = False

        print(
            colorize(
                formatter.format_header('fuzzing session starts', header_line='='),
                AnsiColor.BOLD,
            ),
        )
        print(_get_info())

        self.current_tag = None         # type: Optional[str]

        self.stats = Counter()          # type: Counter
        self.start_time = datetime.now()

        self.warnings = []              # type: List[str]
        self.vulnerable_results = []    # type: List[FuzzingResult]

    def record_result(self, result: FuzzingResult) -> None:
        # Change tag, if appropriate.
        current_tag = result.requests[0].tag
        if current_tag != self.current_tag:
            prefix = '\n' if self.current_tag else ''
            print(f'{prefix}{current_tag} ', end='')

            self.current_tag = current_tag

        # Show status as tests pass/fail.
        is_successful = False
        if (
            len(result.responses) == len(result.requests) and
            not any(result.responses.test_results.values())
        ):
            is_successful = True

        # Record stats for summary output.
        if is_successful:
            print(colorize('.', AnsiColor.LIGHT_GREEN), end='')
            self.stats['success'] += 1
        else:
            print(colorize('E', AnsiColor.RED), end='')
            self.stats['failure'] += 1

            result.log_output = log.stream.getvalue()

            self.vulnerable_results.append(result)

        # Clear log after every run.
        log.clear_stream()

        # Aggregate warnings for summary output.
        self.warnings.append(
            formatter.format_warning(
                f'{result.requests[-1].id}',
                root_logger.stream.getvalue(),
            ),
        )
        root_logger.clear_stream()

    def record_exception(
        self,
        result: FuzzingResult,
        e: BaseException,
    ) -> None:
        # If there's an exception, it means that the first request failed after
        # a sequence of successful requests.
        parameters = result.requests[len(result.responses.responses) - 1].fuzzed_input
        result.exception_info = {
            'name': e.__class__.__name__,
            'traceback': ''.join(
                traceback.format_exception(
                    *sys.exc_info(),
                ),
            ).strip(),
            'parameters': parameters,
        }

    def show_results(self) -> None:
        print()

        print(formatter.format_results(self.vulnerable_results))

        warnings = list(filter(bool, self.warnings))
        if warnings:
            print(formatter.format_warnings(warnings))

        self.stats['warnings'] = len(warnings)
        print(
            formatter.format_summary(
                self.stats,
                datetime.now() - self.start_time,
            ),
        )


def _get_info() -> str:
    return textwrap.dedent(f"""
        Hypothesis Seed: {get_settings().seed}
    """)[1:]



from .supplements import exclude                        # noqa: F401
from .supplements import hooks                          # noqa: F401
from .supplements import include                        # noqa: F401
from .supplements.auth import attacker_account          # noqa: F401
from .supplements.auth import victim_account            # noqa: F401
from .supplements.factory import register_factory       # noqa: F401
from .supplements.request import custom_swagger_client  # noqa: F401
from .supplements.request import make_request           # noqa: F401
from .supplements.setup import setup                    # noqa: F401



import json
import logging
import subprocess
import sys
import textwrap
from datetime import timedelta
from functools import lru_cache
from typing import List
from typing import MutableMapping
from typing import Tuple

from ..result import FuzzingResult
from .color import AnsiColor
from .color import colorize
from .logging import log


def format_results(
    results: List[FuzzingResult],
) -> str:
    """Used to output final results."""
    if not results:
        return ''

    sequences_output = ''
    for result in results:
        name, info = _format_result(result)
        sequences_output += textwrap.dedent("""
            {header}
            {info}
        """)[1:].format(
            header=colorize(
                format_header(
                    '{} [{}]'.format(
                        result.requests[-1].id,
                        name,
                    ),
                    header_line='_',
                ),
                AnsiColor.RED,
            ),
            info=info,
        )

        if not result.log_output:
            continue

        sequences_output += textwrap.dedent("""
            {header}
            {logs}
        """)[1:].format(
            header=format_header(
                'Captured log calls',
                header_line='-',
            ),
            logs=result.log_output,
        )

    return textwrap.dedent("""
        {section_header}
        {sequences}
    """)[1:-1].format(
        section_header=format_header('Test Failures'),
        sequences=sequences_output,
    )


def _format_result(result: FuzzingResult) -> Tuple[str, str]:
    if result.exception_info:
        info = [
            result.exception_info['traceback'],
        ]

        if log.level <= logging.INFO:
            info = [
                'Input:',
                json.dumps(result.exception_info['parameters'], indent=2),
                '',
                *info,
            ]

        return (
            result.exception_info['name'],
            '\n'.join(info),
        )

    return (
        ','.join([
            key
            for key, value in result.responses.test_results.items()
            if value
        ]),
        textwrap.dedent("""
            {request_sequence_header}
            {request_json_dump}
        """)[1:-1].format(
            request_sequence_header=colorize(
                'Request Sequence:',
                AnsiColor.BOLD,
            ),
            request_json_dump=json.dumps(
                [
                    str(request)
                    for request in result.requests
                ],
                indent=2,
            ),
        ),
    )


def format_warnings(warnings: List[str]) -> str:
    """Used to output aggregated warnings."""
    return textwrap.dedent("""
        {header}
        {content}
    """)[1:-1].format(
        header=colorize(
            format_header('warnings summary'),
            AnsiColor.YELLOW,
        ),
        content=''.join(warnings),
    )


def format_summary(
    stats: MutableMapping[str, int],
    timing: timedelta,
) -> str:
    num_passed = stats['success']
    num_failures = stats['failure']
    num_warnings = stats['warnings']

    color = AnsiColor.RESET
    summary = []
    if num_passed:
        summary.append(f'{num_passed} passed')
        color = AnsiColor.LIGHT_GREEN

    if num_warnings:
        summary.append(f'{num_warnings} warnings')
        color = AnsiColor.YELLOW

    if num_failures:
        summary = [f'{num_failures} failed', *summary]
        color = AnsiColor.RED

    if not summary:
        return colorize(format_header('No tests run!'), AnsiColor.BOLD)

    if color == AnsiColor.RESET:
        summary_string = 'No tests run'
    else:
        summary_string = '{} in {} seconds'.format(
            ', '.join(summary),
            round(timing.total_seconds(), 2),
        )

    return colorize(
        colorize(
            format_header(summary_string),
            color,
        ),
        AnsiColor.BOLD,
    )


def format_header(
    message: str,
    header_line: str = '=',
) -> str:
    """
    Creates messages like:
        ========== header text here ==========
    """
    width = _get_terminal_width()

    # Pad the message with single space on both ends, for nicer output.
    message = f' {message} '

    if len(message) >= width:
        return message

    # Add header_line to fill up space.
    padding_length = width - len(message)
    if padding_length % 2 == 1:
        padding_length -= 1

    padding_per_side = padding_length // 2
    return '{}{}{}'.format(
        padding_per_side * header_line,
        message,
        padding_per_side * header_line,
    )


@lru_cache(maxsize=1)
def _get_terminal_width() -> int:  # pragma: no cover
    """
    Source: https://gist.github.com/Steelsouls/9626112
    """
    try:
        width = subprocess.check_output('tput cols'.split())
    except OSError as e:
        print(
            f'Invalid Command: tput: exit status ({e.errno})',
            file=sys.stderr,
        )
    except subprocess.CalledProcessError as e:
        print(
            f'Invalid Command: tput: exit status ({e.returncode})',
            file=sys.stderr,
        )
    else:
        return int(width.decode('utf-8'))

    return 0


def format_warning(
    test_id: str,
    warning: str,
) -> str:
    """
    Formats warning messages for a single endpoint.
    Example Output:
        basic.get_private_listing
          /path/to/file.py:345 Warning: This is a warning message!
    """
    if not warning:
        return ''

    return textwrap.dedent(f"""
        {test_id}
          {warning}
    """)[1:-1]



import os

from setuptools import find_packages
from setuptools import setup


def local_file(path):
    return os.path.relpath(
        os.path.join(
            os.path.dirname(__file__),
            path,
        ),
    )


# We can't import fuzz_lightyear.version, since it pulls in all other
# fuzz_lightyear modules. Therefore, we need to execute the file
# directly.
VERSION = None      # needed for flake8
with open(local_file('fuzz_lightyear/version.py')) as f:
    exec(f.read())


setup(
    name='fuzz_lightyear',
    packages=find_packages(exclude=(['test*', 'tmp*'])),
    version=VERSION,
    description='Vulnerability Discovery through Stateful Swagger Fuzzing',
    license='Copyright Yelp, Inc. 2019',
    author=', '.join([
        'Aaron Loo <aaronloo@yelp.com>',
        'Joey Lee <joeylee@yelp.com>',
        'Victor Zhou <vzhou@yelp.com>',
    ]),
    keywords=[
        'fuzzer',
        'security',
        'swagger',
    ],
    # Remember to update requirements-minimal.txt when making changes
    # to this list.
    install_requires=[
        'bravado',
        'cached-property',
        'hypothesis>=4.56.1',
        'pyyaml',
    ],
    entry_points={
        'console_scripts': [
            'fuzz-lightyear = fuzz_lightyear.main:main',
        ],
    },
    classifiers=[
        'Programming Language :: Python :: 3',
        'Intended Audience :: Developers',
        'Topic :: Security',
        'Topic :: Software Development',
        'Topic :: Software Development :: Testing :: Acceptance',
        'Topic :: Utilities',
        'Environment :: Console',
        'Operating System :: OS Independent',
        'Development Status :: 4 - Beta',
    ],
)



import sys

from .color import AnsiColor
from .color import colorize


def print_error(message: str) -> None:
    print(
        '{} {}'.format(
            colorize('error:', AnsiColor.RED),
            message,
        ),
        file=sys.stderr,
    )


def print_warning(message: str) -> None:
    print(
        '{} {}'.format(
            colorize('warning:', AnsiColor.YELLOW),
            message,
        ),
        file=sys.stderr,
    )



from typing import Any
from typing import List

from bravado.exception import HTTPError
from bravado_core.exception import SwaggerMappingError      # type: ignore
from jsonschema.exceptions import ValidationError           # type: ignore

from ..datastore import get_non_vulnerable_operations
from ..request import FuzzingRequest
from ..response import ResponseSequence
from ..runner import run_sequence
from ..supplements.abstraction import get_abstraction
from .base import BasePlugin


class IDORPlugin(BasePlugin):

    @staticmethod
    def should_run(
        request_sequence: List[FuzzingRequest],
        response_sequence: List[Any],
    ) -> bool:
        last_request = request_sequence[-1]

        # If an operation is marked is non-vulnerable, we don't need
        # to check if it's vulnerable.
        if last_request.operation_id in get_non_vulnerable_operations():
            return False

        # If there are no parameters, then there's no way to
        # specify what victim resources to try to steal / modify.
        return bool(last_request.fuzzed_input)

    @staticmethod
    def is_vulnerable(
        request_sequence: List[FuzzingRequest],
        response_sequence: List[Any],
    ) -> bool:
        run_sequence(
            sequence=request_sequence[:-1],
            responses=ResponseSequence(),
            auth=get_abstraction().get_attacker_session(),  # type: ignore
        )
        try:
            request_sequence[-1].send(
                auth=get_abstraction().get_attacker_session(),  # type: ignore
                should_log=False,
            )
            return True

        except (HTTPError, SwaggerMappingError, ValidationError):
            return False



import logging
from io import StringIO
from typing import Any


class CustomLogger:
    """
    This shims `logging.Logger` without using inheritance, since we want to
    leverage `logging.getLogger`.
    """

    def __init__(
        self,
        log: logging.Logger,
        log_format: str = '%(message)s',
    ) -> None:
        # To prevent infinite recursion, we explicitly call `object`'s setattr.
        # This also sets it as an instance variable, rather than calling the
        # CustomLogger's `__setattr__`, which will set it as a variable on
        # `self.log`.
        super().__setattr__('log', log)
        super().__setattr__('log_format', log_format)

    def set_debug_level(self, debug_level: int) -> None:
        """
        :param debug_level: configure verbosity of log (between 0-2)
        """
        mapping = {
            # Anything over INFO level.
            0: logging.INFO + 1,
            1: logging.INFO,
            2: logging.DEBUG,
        }

        self.setLevel(
            mapping[min(debug_level, 2)],
        )

    def clear_stream(self) -> None:
        """It's easier to create a new stream, rather than clearing it."""
        # Attach stream to log instance, so can be accessible publically.
        self.stream = StringIO()

        handler = logging.StreamHandler(self.stream)
        handler.setFormatter(
            logging.Formatter(self.log_format),
        )

        if not self.handlers:
            self.addHandler(handler)
        else:
            self.handlers[0] = handler

    def __getattr__(self, name: str) -> Any:
        return getattr(self.log, name)

    def __setattr__(self, name: str, value: Any) -> None:
        setattr(self.log, name, value)


def get_logger(
    name: str = None,
    format: str = '%(message)s',
    propagate: bool = True,
) -> CustomLogger:
    """
    :param name: used for declaring log channels.
    """
    log = CustomLogger(
        logging.getLogger(name),
        log_format=format,
    )
    log.propagate = propagate

    log.handlers = []
    log.clear_stream()
    log.set_debug_level(0)

    return log


logging.captureWarnings(True)

# We distinguish the difference between the root logger, and this application
# specific logger. The root logger is for obtaining logged information from
# dependencies. In contrast, the application specific logger is for logging
# information during the normal process of this application.
root_logger = get_logger()
log = get_logger(
    'fuzz_lightyear',
    format='[%(module)s:%(lineno)d]\t%(levelname)s\t%(message)s',

    propagate=False,
)



from typing import Any
from typing import Dict
from typing import List
from typing import Optional

from .datastore import clear_cache
from .datastore import get_user_defined_mapping
from .request import FuzzingRequest
from .response import ResponseSequence


def validate_sequence(
    sequence: List[FuzzingRequest],
    responses: ResponseSequence,
) -> ResponseSequence:

    run_sequence(sequence, responses)
    # Then, check for vulnerabilities.
    responses.analyze_requests(sequence)
    clear_cache()
    return responses


def run_sequence(
    sequence: List[FuzzingRequest],
    responses: ResponseSequence,
    auth: Optional[Dict[str, Any]] = None,
) -> ResponseSequence:

    # First, determine whether this is a successful request sequence.
    for request in sequence:
        response = request.send(
            data=responses.data,
            auth=auth,
        )

        responses.add_response(response)

        # The Response object already saves the output of the response.
        # However, we also still want to save some of the *inputs* to the
        # request, if they were factory-generated. This tries to ensure
        # the same object is referenced throughout the sequence, rather than
        # having the fuzzing engine constantly generate new objects.
        for key in request.fuzzed_input:                            # type: ignore
            if key in get_user_defined_mapping():
                responses.data[key] = request.fuzzed_input[key]     # type: ignore
    return responses



from typing import Callable
from typing import Iterable

from fuzz_lightyear.datastore import get_included_tags


def tags(func: Callable[[], Iterable[str]]) -> Callable:
    """Allows developers to specify Swagger tags which
    should be fuzzed.

    Example:
        Only fuzz operations with the 'user_account' tag.
            >>> @fuzz_lightyear.include.tags
            ... def a():
            ...     return ['user_account']
    """
    tags_to_include = func()
    if tags_to_include:
        get_included_tags().update(tags_to_include)

    return func
